__version__ = "0.2.1"
__author__ = "Arie Bovenberg"
__copyright__ = "2019, Arie Bovenberg"
__license__ = "Apache 2.0"
__description__ = "Capable GraphQL client"

__all__ = [
    "__author__",
    "__version__",
    "__copyright__",
    "__license__",
    "__description__",
]
